@@ -18,6 +18,7 @@
 #include <memory>
 #include <vector>
 
+#include "gmock/gmock-matchers.h"
 #include "gmock/gmock.h"
 #include "gtest/gtest.h"
 
@@ -601,6 +602,58 @@ TEST_F(TestConvertParquetSchema, ParquetLists) {
     arrow_fields.push_back(::arrow::field("name", arrow_list, false));
   }
 
+  // Two-level encoding List<List<Integer>>:
+  // optional group my_list (LIST) {
+  //   repeated group array (LIST) {
+  //     repeated int32 array;
+  //   }
+  // }
+  {
+    auto inner_array =
+        PrimitiveNode::Make("array", Repetition::REPEATED, ParquetType::INT32);
+    auto outer_array = GroupNode::Make("array", Repetition::REPEATED, {inner_array},
+                                       ConvertedType::LIST);
+    parquet_fields.push_back(GroupNode::Make("my_list", Repetition::OPTIONAL,
+                                             {outer_array}, ConvertedType::LIST));
+    auto arrow_inner_array = ::arrow::field("array", INT32, /*nullable=*/false);
+    auto arrow_outer_array =
+        ::arrow::field("array", ::arrow::list(arrow_inner_array), /*nullable=*/false);
+    auto arrow_list = ::arrow::list(arrow_outer_array);
+    arrow_fields.push_back(::arrow::field("my_list", arrow_list, true));
+  }
+
+  // List<Map<String, String>> in three-level list encoding:
+  // optional group my_list (LIST) {
+  //   repeated group list {
+  //     required group element (MAP) {
+  //       repeated group key_value {
+  //         required binary key (STRING);
+  //         optional binary value (STRING);
+  //       }
+  //     }
+  //   }
+  // }
+  {
+    auto key = PrimitiveNode::Make("key", Repetition::REQUIRED, ParquetType::BYTE_ARRAY,
+                                   ConvertedType::UTF8);
+    auto value = PrimitiveNode::Make("value", Repetition::OPTIONAL,
+                                     ParquetType::BYTE_ARRAY, ConvertedType::UTF8);
+    auto key_value = GroupNode::Make("key_value", Repetition::REPEATED, {key, value});
+    auto element =
+        GroupNode::Make("element", Repetition::REQUIRED, {key_value}, ConvertedType::MAP);
+    auto list = GroupNode::Make("list", Repetition::REPEATED, {element});
+    parquet_fields.push_back(
+        GroupNode::Make("my_list", Repetition::OPTIONAL, {list}, ConvertedType::LIST));
+
+    auto arrow_key = ::arrow::field("key", UTF8, /*nullable=*/false);
+    auto arrow_value = ::arrow::field("value", UTF8, /*nullable=*/true);
+    auto arrow_element = ::arrow::field(
+        "element", std::make_shared<::arrow::MapType>(arrow_key, arrow_value),
+        /*nullable=*/false);
+    auto arrow_list = ::arrow::list(arrow_element);
+    arrow_fields.push_back(::arrow::field("my_list", arrow_list, /*nullable=*/true));
+  }
+
   auto arrow_schema = ::arrow::schema(arrow_fields);
   ASSERT_OK(ConvertSchema(parquet_fields));
 
@@ -727,6 +780,81 @@ TEST_F(TestConvertParquetSchema, ParquetRepeatedNestedSchema) {
   ASSERT_NO_FATAL_FAILURE(CheckFlatSchema(arrow_schema));
 }
 
+TEST_F(TestConvertParquetSchema, IllegalParquetNestedSchema) {
+  // List<Map<String, String>> in two-level list encoding:
+  //
+  // optional group my_list (LIST) {
+  //   repeated group array (MAP) {
+  //     repeated group key_value {
+  //       required binary key (STRING);
+  //       optional binary value (STRING);
+  //     }
+  //   }
+  // }
+  {
+    auto key = PrimitiveNode::Make("key", Repetition::REQUIRED, ParquetType::BYTE_ARRAY,
+                                   ConvertedType::UTF8);
+    auto value = PrimitiveNode::Make("value", Repetition::OPTIONAL,
+                                     ParquetType::BYTE_ARRAY, ConvertedType::UTF8);
+    auto key_value = GroupNode::Make("key_value", Repetition::REPEATED, {key, value});
+    auto array =
+        GroupNode::Make("array", Repetition::REPEATED, {key_value}, ConvertedType::MAP);
+    std::vector<NodePtr> parquet_fields;
+    parquet_fields.push_back(
+        GroupNode::Make("my_list", Repetition::OPTIONAL, {array}, ConvertedType::LIST));
+
+    EXPECT_RAISES_WITH_MESSAGE_THAT(
+        Invalid,
+        testing::HasSubstr("Group with one repeated child must be LIST-annotated."),
+        ConvertSchema(parquet_fields));
+  }
+
+  // List<List<String>>: outer list is two-level encoding, inner list is three-level
+  //
+  // optional group my_list (LIST) {
+  //   repeated group array (LIST) {
+  //     repeated group list {
+  //       required binary element (STRING);
+  //     }
+  //   }
+  // }
+  {
+    auto element = PrimitiveNode::Make("element", Repetition::REQUIRED,
+                                       ParquetType::BYTE_ARRAY, ConvertedType::UTF8);
+    auto list = GroupNode::Make("list", Repetition::REPEATED, {element});
+    auto array =
+        GroupNode::Make("array", Repetition::REPEATED, {list}, ConvertedType::LIST);
+    std::vector<NodePtr> parquet_fields;
+    parquet_fields.push_back(
+        GroupNode::Make("my_list", Repetition::OPTIONAL, {array}, ConvertedType::LIST));
+
+    EXPECT_RAISES_WITH_MESSAGE_THAT(
+        Invalid, testing::HasSubstr("LIST-annotated groups must not be repeated."),
+        ConvertSchema(parquet_fields));
+  }
+  // List<List<>>: outer list is two-level encoding, inner list is empty.
+  //
+  // optional group my_list (LIST) {
+  //   repeated group array (LIST) {
+  //     repeated group list {
+  //     }
+  //   }
+  // }
+  {
+    auto list = GroupNode::Make("list", Repetition::REPEATED, {});
+    auto array =
+        GroupNode::Make("array", Repetition::REPEATED, {list}, ConvertedType::LIST);
+    std::vector<NodePtr> parquet_fields;
+    parquet_fields.push_back(
+        GroupNode::Make("my_list", Repetition::OPTIONAL, {array}, ConvertedType::LIST));
+
+    EXPECT_RAISES_WITH_MESSAGE_THAT(
+        Invalid,
+        testing::HasSubstr("LIST-annotated groups must have at least one child."),
+        ConvertSchema(parquet_fields));
+  }
+}
+
 Status ArrowSchemaToParquetMetadata(std::shared_ptr<::arrow::Schema>& arrow_schema,
                                     std::shared_ptr<KeyValueMetadata>& metadata) {
   ARROW_ASSIGN_OR_RAISE(
@@ -1846,7 +1974,9 @@ TEST_F(TestLevels, ListErrors) {
   {
     ::arrow::Status error = MaybeSetParquetSchema(GroupNode::Make(
         "child_list", Repetition::REPEATED,
-        {PrimitiveNode::Make("bool", Repetition::REPEATED, ParquetType::BOOLEAN)},
+        {GroupNode::Make("list", Repetition::REPEATED,
+                         {PrimitiveNode::Make("element", Repetition::REQUIRED,
+                                              ParquetType::BOOLEAN)})},
         LogicalType::List()));
     ASSERT_RAISES(Invalid, error);
     std::string expected("LIST-annotated groups must not be repeated.");
